/*
 *
 *                Copyright 1991,1992,1993 Starbase Corp.
 *                        All rights reserved
 *
 *    This source code is provided to the purchaser of the product
 *    for the purpose of allowing the purchaser to customize and/or
 *    extend the functionality of the product.  The source code   
 *    remains the property of Starbase Corp. with all rights reserved.
 *    Any modifications to the source code are considered derivative
 *    works and all rights thereto are reserved to Starbase Corp. except
 *    that the purchaser may use the modified product in the same
 *    permitted manner as the unmodified product.  It is expressly
 *    prohibited to distribute in any manner any part of the original
 *    source code, whether in source or object form, without the
 *    express written permission from Starbase Corp.
 *
 *    Owner:
 *        Starbase Corporation
 *        1075 NW Murray Blvd., Suite 268
 *        Portland OR 97229
 *
 */

#include <windows.h>
#include <string.h>
#include <ctype.h>
#include "exports.h"
#include "cwstart.h"

#define MAX_BUF_LEN    100

static LPVOID _v_hash_names        = NIL;  /* pointer to the hash table containing */
                                           /* all 'Verilog' keyword names. */
static LPVOID _v_hash_names_1      = NIL;  /* pointer to the hash table containing */
                                           /* user defined keyword names. */
static int _v_hash_size_1          = 47;
static LPVOID _v_hash_names_2      = NIL;  /* pointer to the hash table containing */
                                           /* user defined keyword names. */
static int _v_hash_size_2          = 47;

static LPVOID _v_hash_templates    = NIL;  /* pointer to the hash table containing */
                                           /* an abbreviated name and its template */
static int _v_numlines = 0;

/* Define a structure to hold an abbreviation and the template to be
 * expanded when the abbreviation is entered into the buffer.
 * This structure will be the element in the hash table _v_hash_templates.
 */
typedef struct _LanguageTemplate {         
	LPSTR abbrev;
	LPSTR template;
} LanguageTemplate;


LIBMAIN // defines entry point


/* This string contains all keyword names which will be indented an extra level.
 */
LPSTR VERILOG_INDENT_STR = "\377if\377while\377else\377\377for\377switch\377\
	always\377begin\377task\377begin\377repeat\377module\377function\377case\
	\377initial\377table\377";

/*
 * _v_templates[] contains the keyword abbreviation and the template for the
 * abbreviation.  To add a new name, insert it before the NIL entry or
 * call _v_assign_template().
 *
 * The special characters within the template are:
 *
 *     \n - new line
 *     &  - cursor position after template insertion
 *     @  - backspace
 *     \c - insert 'c' literally
 */
LanguageTemplate _v_templates[] = {
	{ "mod|modu|modul|module",		"module%0%1&\n%2endmodule" },
	{ "beg|begi|begin",				"begin%0%1&\n%2end" },
	{ "cas|case",					"case(&)%0%1\ndefault:\n%2endcase" },
	{ "fork",						"fork%0%1&\n%2join" },
	{ "func|funct|function", 		"function%0%1\n%2endfunction" },
	{ "for",						"for (&; ; )" },
	{ "rep|repe|repea|repeat",		"repeat( & )" },
	{ "tab|tabl|table",				"table%0%1&\n%2endtable" },
	{ "pri|prim|primi|primitive",	"primitive%0%1&\n%2endprimitive" },
	{ "tas|task",					"task%0%1&\n%2endtask" },

	{ NIL,       NIL },
};

/* Define the list of 'Verilog' keyword names to appear in
 * the keyword color if language coloring is enabled.
 */
LPSTR _v_names[] = {

"and",            "for",          "output",      "strong1",
"always",         "force",        "parameter",   "supply0",
"assign",         "forever",      "pmos",        "supply1",
"begin",          "fork",         "posedge",     "table",
"buf",            "function",     "primitive",   "task",
"bufif0",         "highz0",       "pulldown",    "tran",
"bufif1",         "highz1",       "pullup",      "tranif0",
"case",           "if",           "pull0",       "tranif1",
"casex",          "ifnone",       "pull1",       "time",
"casez",          "initial",      "rcmos",       "tri",
"cmos",           "inout",        "real",        "triand",
"deassign",       "input",        "realtime",    "trior",
"default",        "integer",      "reg",         "trireg",
"defparam",       "join",         "release",     "tri0",
"disable",        "large",        "repeat",      "tri1",
"edge",           "macromodule",  "rnmos",       "vectored",
"else",           "medium",       "rpmos",       "wait",
"end",            "module",       "rtran",       "wand",
"endcase",        "nand",         "rtranif0",    "weak0",
"endfunction",    "negedge",      "rtranif1",    "weak1",
"endprimitive",   "nor",          "scalared",    "while",
"endmodule",      "not",          "small",       "wire",
"endspecify",     "notif0",       "specify",     "wor",
"endtable",       "notif1",       "specparam",   "xnor",
"endtask",        "nmos",         "strength",    "xor",
"event",          "or",           "strong0",      	        
                            
	NIL, };

LPSTR _v_sysfuncs[] = {

   "$cleartrace", "$display",  "$finish", "$monitor", "$scope", "$settrace", 
   "$showscopes", "$showvars", "$stop",   "$time",    
	NIL, };



static long _v_in_comment( long offset );
static long _v_find_comment_start( void );
static long _v_find_comment_end( WORD style );

void DLL _v_init( void );
BOOL DLL _v_expand( void );
long DLL _v_language_colors_update( long firstline, long lastline );
void DLL _v_language_colors( long numlines );
void DLL _v_assign_template( LPSTR abbrev, LPSTR template );
void DLL _v_add_keyword( LPSTR keyword, int table );
BOOL DLL _v_indent( void );
BOOL DLL _v_insert_brace( void );

void DLL
_init()
{
	_v_init();
	
}


/*
 ** _v_init()
 *
 *  Initialize this module by exported any required functions and
 *  the hash tables needed to support indenting, template expansion
 *  and keyword coloring.
 */
void DLL
_v_init( void )
{
	static BOOL firstTime = TRUE;
	int    i;

	if (firstTime)
	{
		LibExport( "BOOL _v_expand");
		LibExport( "_v_assign_template LPSTR LPSTR");
		LibExport( "_v_add_keyword LPSTR int");
		LibExport( "BOOL _v_indent");
		LibExport( "BOOL _v_insert_brace");
		LibExport( "long _v_language_colors_update long long");
		LibExport( "_v_language_colors long");
		LibExport( "_v_routines");
		LibExport( "_v_hash_size int int" );

		firstTime = FALSE;

		/* Create a hash table to hold all of the 'Verilog' keyword names.
		 */
		_v_hash_names = HashCreateTable( 41 /* a prime # */, sizeof( LPSTR ), 0 );
		i = 0;

		while ( _v_names[ i ] != NIL )
		{
			LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _v_hash_names, 
				_v_names[i], STRLEN( _v_names[i] ) );
			if (lpf)
			{
				*lpf = StrNew( _v_names[i] );
				i++;
			}
		}

        /* Create a hash table to hold all of the 'verilog' native nodes.
         */
        _v_hash_names_1 = HashCreateTable( 13, sizeof( LPSTR ), 0 );
        i = 0;

        while ( _v_sysfuncs[ i ] != NIL )
        {
            LPSTR XFAR *lpf = (LPSTR XFAR *)HashGetEntry( _v_hash_names_1, 
                    _v_sysfuncs[i], strlen( _v_sysfuncs[i] ) );
            if (lpf)
            {
                *lpf = StrNew( _v_sysfuncs[i] );
                i++;
            }
        }

		/* Create a hash table to hold all of the 'Verilog' template names.
		 */
		_v_hash_templates = HashCreateTable( 17 /* a prime # */, sizeof( LPSTR ), 0 );
		i = 0;

		while ( _v_templates[ i ].abbrev != NIL )
		{
			_v_assign_template( _v_templates[i].abbrev, _v_templates[i].template );
			i++;
		}
		CompilerNewExt( ".v" );
	}
}


void DLL
_v_hash_size( int ht, int size )
{
	if (size > 0)
	{
		if (ht == 1)
			_v_hash_size_1 = size;
		else if (ht == 2)
			_v_hash_size_2 = size;
	}
}

PUBLIC BOOL
_v_in_string( long offset )
{
	UINT  ch;
	long  quotes     = 0;

	/* Check to see if we are currently positioned inside a 'C' string
	 */
	_PosInit( offset );
	ch = _PosCurrentChar();

	while (TRUE)
	{
		do
		{
			if (ch == '"')
			{
				//ch = _PosPrevChar();
				quotes++;
			}
			ch = _PosPrevChar();
		} while ((ch != EOF_CHAR) && (ch != '\n'));

		if (ch == '\n')
			ch = _PosPrevChar();
	 	if (ch == '\r')
			ch = _PosPrevChar();
		if (ch != '\\')
			break;
		else
			ch = _PosPrevChar();
	}
	return ((int)quotes & 1);
}

/*
 ** _v_expand()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'Verilog' language template expansion.  This function
 *    is normally bound to the <SPACE> key.  When invoked, it reads the 
 *    previous word.  If the word is in the list of defined templates,
 *    the template is extracted and inserted.  Some characters in the 
 *    templates have special meanings. They are:
 *
 *       \n - new line
 *       &  - cursor position after template insertion
 *       @  - backspace
 *       \  - take next character literally
 */
BOOL DLL
_v_expand( void )
{
	LPVOID htEntry;
	LPSTR  _v_string = NULL;
	LPSTR  str;
	int    i = 0;

	MarkSavePos();

	/* Check to see if we are currently positioned inside a 'Verilog' comment
	 */
	if ((_v_in_comment( -1L ) != -1)
			|| _v_in_string(-1))
		goto _v_expand_incomplete;

	str = BufReadStr( 65535 );
	while (str[i] && !ISSPACE(str[i]))
		i++;
	str[i] = '\0';

	_v_string = StrTrim( str, " \t" );
	StrFree( str );

	if (!_v_string || !*_v_string)
		goto _v_expand_incomplete;

	/* Search for a keyword match in the hash table.
	 */
	htEntry = HashFindEntry( _v_hash_templates, _v_string, STRLEN( _v_string ) );

	if (htEntry)
	{
		/* A template exists for the current keyword, delete the keyword
		 * and insert the template.
		 */
		MarkDropPos( 1 );
		str = *(LPSTR*)htEntry;
		BufDelChar( lstrlen( _v_string ) );

		_ext_expand_template( str );

	} 
	else
		goto _v_expand_incomplete;
	StrFree( _v_string );
	return TRUE;

_v_expand_incomplete:
	StrFree( _v_string );
	MarkRestorePos();

	return FALSE;
}


/*
 ** _v_language_colors_update()
 *
 *  This function does an intermediate update of the colors within
 *  a range of line numbers (firstline and lastline).  It must also
 *  determine if the first and last line specified are within a
 *  comment and act appropriately.
 */
long DLL
_v_language_colors_update( long firstline, long lastline )
{
	long startoff;
	long ec_line  = 0; /* end comment line */
	WORD style    = 0;
	int  color;
	long count;

	if (firstline && lastline)
	{
		/* Begin coloring at the previous comment if
		 * currently located in the middle of one.
		 */
		_PosInit( 0 );
		if (firstline == 1)
			_PosPrevLine( 0L );
		else
			_PosNextLine( firstline - 1);

		startoff = _PosQOffset();
		if ( _v_find_comment_start() == 1)
		{
			startoff  = _PosQOffset();

			style    = 1;

			_PosNextChar();   /* move off of the comment */
			_PosNextChar();

			if (_v_find_comment_end(style))
				ec_line = _PosQLine();
		}

		_PosInit( startoff );
		_PosCurrentCharNL();
		while (_PosPrevCharNL() && (_PosPrevCharNL() == '\\'))
		{
			_PosPrevLine( 0L );
			_PosCurrentCharNL();
			startoff  = _PosQOffset();
		}

		_PosInit( startoff );
		firstline = _PosQLine();


		/* End coloring at the next comment if
		 * currently located in the middle of one.
		 */
		_PosInit( 0 );
		color = _PosQColor();
		_PosNextLine( lastline - 1 );
		_PosNextLine( 0 );

		if (color || _PosQColor())
		{
			_v_find_comment_end(style);
			lastline = _PosQLine();
		}
		else if (_v_find_comment_end(style) == 2)
			lastline = _PosQLine();


		SysSetSystemFlags( SYSTEM_DELAY_ATTR_CHANGES );

		if (!ec_line)
			AttrSetColor( firstline, lastline, 1L, 0x7FFFFFFFL, 0 );
		else if (ec_line < lastline)
		{
			AttrSetColor( ec_line, lastline, 1L, 0x7FFFFFFFL, 0 );
			if (ec_line < firstline)
				firstline = ec_line;
		}

		/* Reset the attributes for the line range within _c_language_colors.
		 */
		_PosInit( startoff );
		_v_language_colors( lastline - firstline + 1 );

		/* Allow changes to be redrawn 
		 */
		SysClrSystemFlags( SYSTEM_DELAY_ATTR_CHANGES );
		WinRedrawLines( WinQCurrentWindow(), firstline, 
			(_v_numlines > 0) ? firstline + _v_numlines : lastline );
	}
	count = (lastline - firstline);
	return(count);
}

/*
 ** _v_language_colors()
 *
 * If numlines == -1, start from the top of the buffer and encode all of them,
 * otherwise start at the current location of the number of lines specified.
 *
 * It is assumed that if numlines != -1, that _PosInit() has been called
 * and numlines is relative to the current position.
 */
void DLL
_v_language_colors( long numlines )
{
	long  startoff;
	long  startline;
	long  endoff;
	long  inc;
	WORD  ch;
	LPSTR buf;
	int   i = 0;
	int   comment_color = ColorComments( -1 );
	int   keyword_color = ColorKeywords( -1 );
	long  maxlines = ExtCommentSearchLimit( -1 );
	int   quote_color      = ColorAlternate1( -1 );
	int   number_color     = ColorAlternate2( -1 );
	int   brace_color      = ColorAlternate3( -1 );
	int   preproc_color    = ColorAlternate4( -1 );
	int   doc_comment_color= ColorAlternate7( -1 );


	_v_numlines = 0;
	buf = (LPSTR)MemAlloc( MAX_BUF_LEN );

	/* Start at the top of the buffer or current position 
	 */
	if (numlines == -1)
	{
		_PosInit( 0 );  
		numlines = 0x7FFFFFFF;
	}
	else
	{
		startoff  = _PosQOffset();
		startline = _PosQLine();
		endoff    = _v_in_comment( startoff );
		if (endoff == -1)
			_PosInit( startoff );
		else
		{
			_PosInit( endoff );
			numlines += (startline - _PosQLine());
		}
	}
	ch = _PosCurrentChar();
	startoff = _PosQOffset();

	maxlines = max( numlines, maxlines );

	while ((ch != EOF_CHAR) && (numlines > 0))
	{
		if (ch == '\n')
			numlines--, _v_numlines++;

		if (ch == '"')
		{
			UINT quote = ch;

            startoff = _PosQOffset();

			do 
			{
				ch = _PosNextChar();
				if (ch == '\\')
				{
					ch = _PosNextChar();
					if ((ch == quote) || (ch == '\\') || (ch == '\n') || (ch == '\r'))
					{
						if (ch == '\n')
							numlines++, _v_numlines--;
						else if (ch == '\r')
							ch = _PosNextChar();
						ch = _PosNextChar();
					}
				}
				if ((ch == EOF_CHAR) || (ch == '\n'))
					break;
			} while (ch != quote);

			if (ch == quote)
				ch = _PosNextChar();

			/* color the string */
			if (quote_color)
			{
				endoff = _PosQOffset();
	    		_PosInit( startoff );
				_PosSetColor( quote_color, endoff - startoff );
				_PosInit( endoff );
			}
			continue;
		}
		else if (number_color && (ISDIGIT( ch ) || ch == '\'' ))
		{
            int j = 0;
			UINT lastch = _PosPrevChar();

			if (lastch != EOF_CHAR)
			  	_PosNextChar();

			startoff = _PosQOffset();
			do
			{
                j++;
				ch = _PosNextChar();
			} while (ISXDIGIT( ch ) || (ch == 'o') || (ch == 'b') || (ch == '\'')
					 || (ch == 'h') || (ch == '?') || (ch == '_') );

			if ((  lastch != ' ' 
				&& lastch != '\t' 
				&& lastch != '\r' 
				&& lastch != '\n' 
				&& lastch != EOF_CHAR
				&& !ISPUNCT( lastch ))
				|| lastch == '_'
				|| (ch == '_'))
				continue;

            if ( ch == 'x' )
                ch = _PosNextChar();

			endoff = _PosQOffset();
    		_PosInit( startoff );
			_PosSetColor( number_color, endoff - startoff );
			_PosInit( endoff );
			continue;
		}
		else if (ISALNUM( ch ) || (ch == '_') || (ch == '$'))
		{
			UINT lastch = _PosPrevChar();

			if (lastch != EOF_CHAR)
				_PosNextChar();

			i = 0;
			startoff = _PosQOffset();

			do
			{
				if (i < MAX_BUF_LEN)
					buf[i++] = (char)ch;
				ch = _PosNextChar();
			} while (ISALNUM( ch ) || (ch == '_'));

			if ((lastch != ' ' 
						&& lastch != '\t' 
						&& lastch != '\r' 
						&& lastch != '\n' 
						&& lastch != EOF_CHAR
						&& !ISPUNCT( lastch ))
					|| (ch == '_'))
				i = 0;
			continue;
		}
		else if (i)
		{
			int color = 0;
			if (_v_hash_names_1 && HashFindEntry( _v_hash_names_1, buf, i ))
				color = ColorAlternate5( -1 );
			else if (_v_hash_names_2 && HashFindEntry( _v_hash_names_2, buf, i ))
				color = ColorAlternate6( -1 );
			else if (HashFindEntry( _v_hash_names, buf, i ))
				color = keyword_color;

			if (color)
			{
				endoff = _PosQOffset();
				_PosInit( startoff );
				_PosSetColor( color, endoff - startoff );
				_PosInit( endoff );
			}
			i = 0;
		} 

		if (ch == '/')
		{
			startoff = _PosQOffset();
			ch = _PosNextChar();

			if (ch == '/')
			{
				inc = _PosNextLine( 1 );
				numlines--;
				_v_numlines++;
				endoff = _PosQOffset();
				_PosInit( startoff );
				_PosSetColor( comment_color, endoff - startoff - inc );
				_PosInit( endoff );
				ch = _PosCurrentChar();
			}
			else if (ch == '*')
			{
				long lines = 0;

				ch = _PosNextChar();
				while ((ch != EOF_CHAR) && (lines < maxlines))
				{
					if (ch == '\n')
					{
						lines++;
						numlines--;
						_v_numlines++;
					}
					else if (ch == '*')
					{
						ch = _PosNextChar();
						if (ch == '/')
						{
							endoff = _PosQOffset();
							_PosInit( startoff );
							_PosSetColor( comment_color, endoff - startoff + 1 );
							_PosInit( endoff );
							break;
						}
						else
							continue;
					}
					ch = _PosNextChar();
				}
				ch = _PosNextChar();
			}
			continue;
		}
#ifdef DBC_ABLE
		else if (IsDBChar(ch))
			;
#endif
		else if (brace_color && STRCHR("#", ch) != 0)    
		{                                        
			startoff = _PosQOffset();
			ch = _PosNextChar();

			if (ch == '(')
			{
	            int j = 0;
				UINT lastch = _PosPrevChar();

				//if (lastch != EOF_CHAR)
				  //	_PosNextChar();

				do
				{
	                j++;
					ch = _PosNextChar();
					if (ch == EOF_CHAR)
						break;
				} while (ch != ')');

				ch = _PosNextChar();

				if ((  lastch != ' ' 
					&& lastch != '\t' 
					&& lastch != '\r' 
					&& lastch != '\n' 
					&& lastch != EOF_CHAR
					&& !ISPUNCT( lastch ))
					|| lastch == '_'
					|| (ch == '_'))
					continue;

				endoff = _PosQOffset();
	    		_PosInit( startoff );
				_PosSetColor( brace_color, endoff - startoff );
				_PosInit( endoff );
				continue;
			}
			else
			{
	            int j = 0;
				UINT lastch = _PosPrevChar();

				//if (lastch != EOF_CHAR)
				  //	_PosNextChar();

				do
				{
	                j++;
					ch = _PosNextChar();
				} while (ISALNUM( ch ) || (ch == '.') 	|| (j == 1));

				if ((  lastch != ' ' 
					&& lastch != '\t' 
					&& lastch != '\r' 
					&& lastch != '\n' 
					&& lastch != EOF_CHAR
					&& !ISPUNCT( lastch ))
					|| lastch == '_'
					|| (ch == '_'))
					continue;

				endoff = _PosQOffset();
	    		_PosInit( startoff );
				_PosSetColor( brace_color, endoff - startoff );
				_PosInit( endoff );
				continue;
			}
		}
		else if (preproc_color && STRCHR("@", ch) != 0)
		{
		   	startoff = _PosQOffset();
			ch = _PosNextChar();

		   if (ch == '(')
		   {
	        int j = 0;
		   	UINT lastch = _PosPrevChar();
		   
		   	//if (lastch != EOF_CHAR)
		   	  //	_PosNextChar();
		   
		   	//startoff = _PosQOffset();
		   	//do
		   	//{
	        //       j++;
		   	//	ch = _PosNextChar();
		   	//	if (ch == EOF_CHAR)
		   	//		break;
		   	//} while (ch != ')');
		   
		   	ch = _PosNextChar();
		   
		   	if ((  lastch != ' ' 
		   		&& lastch != '\t' 
		   		&& lastch != '\r' 
		   		&& lastch != '\n' 
		   		&& lastch != EOF_CHAR
		   		&& !ISPUNCT( lastch ))
		   		|| lastch == '_'
		   		|| (ch == '_'))
		   		continue;
		   
		   	endoff = _PosQOffset();
	       	_PosInit( startoff );
		   	//_PosSetColor( preproc_color, endoff - startoff );
			_PosSetColor( keyword_color, 1 );
		   	_PosInit( endoff );
		   	continue;
		   }
		   else
		   {
	       		int j = 0;
				UINT lastch = _PosPrevChar();

     

				//if (lastch != EOF_CHAR)
				  //	_PosNextChar();

				//startoff = _PosQOffset();
		   		do
		   		{
	                j++;
		   			ch = _PosNextChar();
		   		} while (ISALNUM( ch ) || (ch == '.') 	|| (j == 1));
		   
		   		if ((  lastch != ' ' 
		   			&& lastch != '\t' 
		   			&& lastch != '\r' 
		   			&& lastch != '\n' 
		   			&& lastch != EOF_CHAR
		   			&& !ISPUNCT( lastch ))
		   			|| lastch == '_'
		   			|| (ch == '_'))
		   			continue;

				endoff = _PosQOffset();
	    		_PosInit( startoff );
			  //	_PosSetColor( preproc_color, endoff - startoff );
			  _PosSetColor( keyword_color, 1 );
				_PosInit( endoff );
				continue;
			}
		}

		ch = _PosNextChar();
	}
	MemFree( buf );
}


void DLL
_v_add_keyword( LPSTR keyword, int table )
{
	if (keyword && *keyword && table >= 0)
	{
		LPSTR XFAR *lpf;
		LPVOID  ht;

		switch( table )
		{
		case 0:
			ht = _v_hash_names;
			break;
		case 1:
			if (_v_hash_names_1 == NULL)
				_v_hash_names_1 = HashCreateTable( _v_hash_size_1, sizeof( LPSTR ), 0 );
			ht = _v_hash_names_1;
			break;
		case 2:
			if (_v_hash_names_2 == NULL)
				_v_hash_names_2 = HashCreateTable( _v_hash_size_2, sizeof( LPSTR ), 0 );
			ht = _v_hash_names_2;
			break;
		}

		lpf = (LPSTR XFAR *)HashGetEntry( ht, 
								keyword, STRLEN( keyword ) );
		if (lpf)
			*lpf = StrNew( keyword );
	}
}


/*
 ** _v_assign_template()
 *
 *  PARAMETERS:
 *    abbrev   - the keyword abreviation used to locate a template
 *    template - the encoded template string associated with 'abbrev'
 *
 *  DESCRIPTION:
 *    Provide a means of adding or modifying 'Verilog' template strings
 *    without changing the code.
 */
void DLL
_v_assign_template( LPSTR abbrev, LPSTR template )
{
	_ext_assign_template( _v_hash_templates, abbrev, template );
}

/*
 ** _v_indent()
 *
 *  DESCRIPTION:
 *    Attempts to perform 'Verilog' language smart indenting.  This function
 *    will behave differently based on the following conditions:
 *
 *    1. Within a comment - normal indentation
 *
 *    2. Not positioned at end of line - normal indentation
 *
 *    3. Current line prefixed by a name in JAV_INDENT_STR - perform smart
 *       indentation.  This may include skipping the next line if
 *       it consists only of an open brace.
 */
BOOL DLL
_v_indent( void )
{
    long   curline;
    long   col;
    LPSTR  _v_string;
    LPSTR  str;
    long   matchpos;

	MarkSavePos();

	curline = BufQCurrentLine();

	/* Check to see if we are currently positioned inside a 'Verilog' comment
	 */
	if (_v_in_comment( -1L ) != -1)
		goto _v_indent_incomplete;

    /* Not in a comment so indent according to the current 'VRML' construct.
     */
    if (!SrchFind( "^[ \t]*[a-zA-Z_]+\\c([ \t]|$)", SEARCH_REGEX , NIL )
            || (BufQCurrentLine() != curline) )
        goto _v_indent_incomplete;

    /* Read the keyword at the beginning of the current line.
     */
    col = BufQCurrentCol();
    SrchFind( "(^|[^a-zA-Z_])\\c[a-zA-Z_]", SEARCH_REGEX , NIL );
    str = BufReadStr( col - BufQCurrentCol() );
    _v_string = StrTrim( str, " \t" );
    StrFree( str );

    /* Set prefix and suffix to \377 to perform a quick lookup of the name
     */
    str = StrNew( "\377" );
    _v_string = StrAppend( _v_string, str );
    str = StrAppend( str, _v_string );
    StrFree( _v_string );
    
    matchpos = StrMatch( str, VERILOG_INDENT_STR, SEARCH_IGCASE | SEARCH_FORWARD, NIL );
    StrFree( str );

    if (!matchpos)
        /* Not a 'VRML' keyword, do normal indentation. */
        goto _v_indent_incomplete;
        
    MarkRestorePos();

    _ext_insert_indented_EOL();
    BufInsertChar( '\t' );
	return TRUE;

_v_indent_incomplete:
		MarkRestorePos();

	return FALSE;
}

static LPSTR 
ReadLine(void)
{
	LPSTR pLine, p;
	long  colFirst = BufQCurrentCol();

	pLine = StrNNew("", (int)colFirst);
	_fmemset(pLine, ' ', (int)colFirst);
	pLine[colFirst] = 0;
	pLine = StrAppend(pLine, p = BufReadStr(32767));
	StrFree(p);
	return(pLine);
}
static int 
firstChar(LPSTR str,long pos)
{
	/*
	**	Desc: Returns the first non-white character of the string
	**			before position pos or '\0'if none
	**	 Pre: Input string
	**	Post: first char
	** History:
	**	01-Oct-1990 JVV: Created
	*/

	LPSTR pTmp;
	LPSTR endpoint;


	pTmp = str;
	endpoint = str + pos;
	while ( *pTmp && pTmp < endpoint && (*pTmp == ' ' || *pTmp == '\t'))
		pTmp++;

	if (!(*pTmp) || pTmp == endpoint)
		return 0;

	return(*pTmp);
}

/**************************************************************************/

static int 
lastChar(LPSTR str,long pos)
{
	/*
	**	Desc: Finds the last non whitespace character before the position pos
	**	 Pre: Input string and position
	**	Post: last nonblank char
	** History:
	**	01-Oct-1990 JVV: Created
	*/

	LPSTR pTmp;

	pTmp = str + min(((int)pos),(int)(STRLEN(str) - 1));

	while (pTmp >= str && (*pTmp == ' ' || *pTmp == '\t'))
		pTmp--;

	if (pTmp < str)
		return 0;

	return *pTmp;
}

BOOL DLL
_v_insert_brace( void )
{
	LPVOID htEntry;
	int fEmptyLine;
	LPSTR buf;
	long xTarg, yTarg;

	xTarg = BufQCurrentCol();
	yTarg = BufQCurrentLine();

	if ((_v_in_comment(-1) == -1) &&
		(!_v_in_string(-1)))
	{

		if (xTarg == 1)
		{ /* function delimiter */
			_ext_expand_template( "{\n%1&\n%home}" );
		}
		else
		{
			MovAbs(0, 1);
			buf = ReadLine();
			fEmptyLine = (firstChar(buf, STRLEN(buf)) == 0);
			StrFree(buf);
			MovAbs(yTarg, xTarg);
		
			if (fEmptyLine)
				if (MovUp( 1 ))
					MovEOL();
		
			_v_init();
			htEntry = HashFindEntry( _v_hash_templates, "{", 1 );

			if (htEntry)
				_ext_expand_template( *(LPSTR*)htEntry );
			else
				BufInsertChar( 0x7b );
		}
	}
	else
		BufInsertChar( 0x7b );

	return TRUE;
}

/*--  Private functions --*/


/*
 ** _v_in_comment()
 *
 *  Determines whether or not the current position in the buffer is
 *  located within a 'Verilog' comment by searching backward for a // or /* 
 *  comment declaration.
 *
 *  Returns -1 if not in a comment 
 *          >= 0 indicates the offset in the buffer where the comment begins
 */
static long
_v_in_comment( long offset )
{
	WORD  ch;
	long  lines     = 0;
	long  maxlines = ExtCommentSearchLimit( -1 );

	/* Check to see if we are currently positioned inside a 'Verilog' comment
	 */
	_PosInit( offset );
	ch = _PosCurrentChar();

	do
	{
		switch( ch )
		{
		case '*':
			ch = _PosPrevChar();
			if (ch == '/')    /* Found the beginning of a comment */
			{
				offset = _PosQOffset();  /* currrent line is in a comment */
				ch = _PosPrevChar();
				if ((ch != '/') || (ch == '/' && (lines==0)))
				{
					_PosInit( offset );
					return offset;
				}
			}
			break;
		case '/':
			ch = _PosPrevChar();
			if (ch == '/')
			{
				if (lines==0)
					return _PosQOffset();  /* currrent line is in a comment */
				else
					return -1;    /* Not in a comment */
			}
			else if (ch == '*')
				return -1L;    /* Not in a comment */
			break;

		case '"':
			ch = _PosPrevChar();
			while (TRUE)
			{
				while (ch != EOF_CHAR && ch != '"' && ch != '\n')
					ch = _PosPrevChar();

				if (ch == '\n')
				{
					lines++;
					ch = _PosPrevChar();
				}
			 	if (ch == '\r')
					ch = _PosPrevChar();
				if (ch != '\\')
					break;
				else
					ch = _PosPrevChar();
			}
			ch = _PosPrevChar();
			break;

		case '\n':
			lines++;
		default:
			ch = _PosPrevChar();
			break;
		}
	} while ((ch != EOF_CHAR) && (lines < maxlines));

	return -1L;
}

/*
 ** _v_find_comment_start()
 *
 *  Search backwards for a 'Verilog' comment sequence. Return 1 if /<*> found
 *  and 2 if <*>/ found, 0 otherwise. Leave the current position
 *  at the beginning of the found comment.
 */
static long 
_v_find_comment_start( void )
{
	WORD ch = _PosCurrentChar();
	long  lines = 0;
	long  maxlines = ExtCommentSearchLimit( -1 );

	do
	{
		if (ch == '*')
		{
			ch = _PosPrevChar();
			if (ch == '/')
			{
				long offset = _PosQOffset();  /* currrent line is in a comment */
				ch = _PosPrevChar();
				_PosInit( offset );

				if (ch == '/')
				{
					if (lines == 0)
						break;
				}
				else
					return 1;
			}
		}
		else if (ch == '/')
		{
			ch = _PosPrevChar();
			if ((ch == '/') && (lines == 0))
				break;
			if (ch == '*')
			{
				long offset = _PosQOffset();  /* currrent line is in a comment */
				ch = _PosPrevChar();
				if (ch == '/')
					return 1;
				_PosInit( offset );
				return 2;
			}
		}
		else
		{
			if (ch == '\n')
				lines++;
			else if (ch == '"')
			{
				ch = _PosPrevChar();
				while (TRUE)
				{
					while (ch != EOF_CHAR && ch != '"' && (ch == '\n'))
					{
						if (ch == '\n')
							lines++;
						ch = _PosPrevChar();
					}
					if (ch == '\n')
						ch = _PosPrevChar();
				 	if (ch == '\r')
						ch = _PosPrevChar();
					if (ch != '\\')
						break;
					else
						ch = _PosPrevChar();
				}
			}
			ch = _PosPrevChar();
		}
	} while ((ch != EOF_CHAR) && (lines < maxlines));

	return 0;
}

/*
 ** _v_find_comment_end()
 *
 *  Search forwards for a 'Verilog' comment sequence. Return 1 if /<*> found
 *  and 2 if <*>/ found, 0 otherwise.  Leave the current position
 *  at the end of the found comment.
 */
static long
_v_find_comment_end( WORD style )
{
	long lines = 0;
	UINT  ch = _PosCurrentChar();
	long  maxlines = ExtCommentSearchLimit( -1 );

	do
	{
		if (ch == '*')
		{
			ch = _PosNextChar();
			if (ch == '/')
				return 2;
		}
		else if (ch == '/')
		{
			ch = _PosNextChar();
//			if ((ch == '/') && style)
//				break;
			if (ch == '*')
				return 1;
		}
		else
		{
			if (ch == '\n')
				lines++;
			else if (ch == '"')
			{
				int eolchars;

				ch = _PosNextChar();
				while (TRUE)
				{
					while (ch != EOF_CHAR && ch != '"' && ch != '\n')
					{
						if (ch == '\n')
							lines++;
						ch = _PosNextChar();
					}
					eolchars = 1;
					if (ch == '\n')
						ch = _PosPrevChar();
				 	if (ch == '\r')
						ch = _PosPrevChar(), eolchars = 2;
					if (ch != '\\')
						break;
					else
					{
						_PosNextChar();
						if (eolchars == 2)
							_PosNextChar();
						ch = _PosNextChar();
					}
				}
			}
			ch = _PosNextChar();
		}
	} while ((ch != EOF_CHAR) && (lines < maxlines));

	return 0;
}



void DLL
_v_routines( void )
{
	int   matches = 0;
	LPSTR str     = "^module";
	LPSTR stre    = "^module[ \\t].*(.|\n)*[ \t\n]*\\c[^ \t\n]";
	long  curline;
	UINT  ch;
	long  lines;

	SysWaitStart();

	MarkSavePos();

	BufSetAscii();
	MovTopBuf();

	while (SrchFind( str, SEARCH_REGEX | SEARCH_FORWARD, 0 ))
	{
		curline = BufQCurrentLine();

		SrchFind( stre, SEARCH_REGEX | SEARCH_FORWARD, 0 );

		ch = BufReadChar();
		if (ch != ';')
		{
			if (!matches)
				AttrSetVisible( 0, 0, 0 );
			matches++;
			lines = BufQCurrentLine() - curline
						- !!(ch == '{');
			AttrSetVisible( curline, curline + (lines < 0L ? 0L : lines), 1 );
			MovHome();

			while (SrchFind("(^endmodule|always|[^d]t\\cask)", SEARCH_REGEX | SEARCH_FORWARD, 0 ))
			{
				curline = BufQCurrentLine();
				if (BufQCurrentCol() == 1)
					goto module_end;

				if (!matches)
					AttrSetVisible( 0, 0, 0 );
				matches++;
				lines = BufQCurrentLine() - curline
							- !!(ch == '{');
				AttrSetVisible( curline, curline, 1 );
				MovHome();
				MovDown( 1 );
			}
		}

module_end:

		if (!MovDown( 1 ))
			break; 
	}

	if (matches)
	{
		BufSetCompact();
		MarkDropPos( 1 );
		MovTopBuf();
	}
	else
		MarkRestorePos();

	SysWaitEnd();
	// display message indicating result
	if (matches)
		MsgMessage("");
	else
		MsgMessage("No routines found");
}

